# We estimate the total using a capture and recapture plan: In a population with
# an unknown total N, capture M units without replacement, tag and release them; 
# then capture n units without replacement and record X, the number of tagged units.
 
# X--hypergeometric with N units, M of them are good (N-M bad), select 
# n from N units, there are x good units in the sample.
# P(X=x)=dhyper(x,m,N-m,n)  for max(0,n+M-N)<=x <=min(n,M)
# Here, n and M are known, and want to estimate N based on x.
# i) one-sided 1-alpha CI for N, [L(X),infty)
# ii) one-sided 1-alpha CI for N, [M,U(x)]
# iii) two-sided 1-alpah CI for N, [L(X),U(X)]
# the two-sided interval is the intersection of two one-sided 1-alpha/2 intervals

# phyper(x,a,b,n)=cdf of hypergeometric distribution with
# N=a+b units in the population, a good units, b bad units
# x good units in a sample of size n 

# see #### for a quick way to compute the minimum coverage probability
# use the bisection method and is as quick as hyper-total-improve-for.r
# use the while loop


tim=proc.time()

#alpha=0.05
#M=10          #M>=n ok, M<n not ok 
#n=8

alpha=0.05
M=148
n=82

#M=200
#n=200

# indicator fucntion of interval [a,b]
ind<- function(x,a,b){
(x>=a)*(x<=b)
}

ffnl<-function(N){1-phyper(0,M,N-M,n)} # this function provides an upper bound for the lower limit for N
ii=max(M,n)
ffnL=ffnl(ii)
while(ffnL>1-alpha){
ffnL=ffnl(ii)
ii=ii+1}
Nl=ii            # L(X)<=Nl

lcit<- function(x,n,alpha){   # the lower limit of 1-alpha CP interval
kk=1:length(x)
for (i in kk){
if (x[i]>min(n,M)-0.5){kk[i]=max(n,M)} else {
      aa=max(M,n):Nl
      bb=aa*0+1
      bb[2:length(aa)]=1-phyper(x[i],M,aa[2:length(aa)]-M-1,n)
      #x=15
      #bb[2:length(aa)]=1-phyper(x,M,aa[2:length(aa)]-M,n)
      dd=cbind(aa,bb)
      dd=dd[which(dd[,2]>=1-alpha),]
      if (length(dd)==2){kk[i]=dd[1]} else {kk[i]=max(dd[,1])}
      }}
kk
}

ffnu<-function(N){phyper(0,M,N-M+1,n)} # this function provides an upper bound for the upper limit for N
ii=max(M,n)
ffnU=ffnu(ii)
while(ffnU<1-alpha){
ffnU=ffnu(ii)
ii=ii+1}
Nu=ii            # U(X)<=NU for X>0

ucit<- function(x,n,alpha){   # the lower limit of 1-alpha CP interval
kk=1:length(x)
for (i in kk){
if (x[i]<0.5){kk[i]=Inf} else {
      aa=max(M,n):(Nu+1)
      #aa=M:30
      bb=aa*0
      bb[1:length(aa)]=phyper(x[i]-1,M,aa[1:length(aa)]-M+1,n)
      #x=19
      #bb[1:length(aa)]=phyper(x-1,M,aa[1:length(aa)]-M+1,n)
      dd=cbind(aa,bb)
      #dd=dd[which(dd[,2]>=1-alpha-10^(-15)),]
      dd=dd[which(dd[,2]>=1-alpha),]
      if (length(dd)==2){kk[i]=dd[1]} else {kk[i]=min(dd[,1])}
      }}
kk
}


ffnl2<-function(N){1-phyper(0,M,N-M,n)} # this function provides an upper bound for the lower limit for N
ii=max(n,M)
ffnL2=ffnl2(ii)
while(ffnL2>1-alpha/2){
ffnL2=ffnl2(ii)
ii=ii+1}
Nl2=ii            # L(X)<=Nl2

lcit2<- function(x,n,alpha){   # the lower limit of 1-alpha CP interval
kk=1:length(x)
for (i in kk){
if (x[i]>min(n,M)-0.5){kk[i]=max(n,M)} else {
      aa=max(M,n):Nl2
      bb=aa*0+1
      bb[2:length(aa)]=1-phyper(x[i],M,aa[2:length(aa)]-M-1,n)
      #x=15
      #bb[2:length(aa)]=1-phyper(x,M,aa[2:length(aa)]-M,n)
      dd=cbind(aa,bb)
      dd=dd[which(dd[,2]>=1-alpha/2),]
      if (length(dd)==2){kk[i]=dd[1]} else {kk[i]=max(dd[,1])}
      }}
kk
}

ffnu2<-function(N){phyper(0,M,N-M+1,n)} # this function provides an upper bound for the upper limit for N
ii=max(n,M)
ffnU2=ffnu2(ii)
while(ffnU2<1-alpha/2){
ffnU2=ffnu2(ii)
ii=ii+1}
Nu2=ii            # U(X)<=NU2 for X>0

ucit2<- function(x,n,alpha){   # the lower limit of 1-alpha CP interval
kk=1:length(x)
for (i in kk){
if (x[i]<0.5){kk[i]=Inf} else {
      aa=max(M,n):(Nu2+1)
      #aa=M:30
      bb=aa*0
      bb[1:length(aa)]=phyper(x[i]-1,M,aa[1:length(aa)]-M+1,n)
      #x=19
      #bb[1:length(aa)]=phyper(x-1,M,aa[1:length(aa)]-M+1,n)
      dd=cbind(aa,bb)
      dd=dd[which(dd[,2]>=1-alpha/2-10^(-15)),]
      if (length(dd)==2){kk[i]=dd[1]} else {kk[i]=min(dd[,1])}
      }}
kk
}

xx=0:min(n,M)

xx=0:min(n,M)
lcin2=lcit(xx,n,alpha)   # this is the lower one-sided 1-alpha interval 
ucin2=ucit(xx,n,alpha)   # this is the upper one-sided 1-alpha interval
lcin1=lcit2(xx,n,alpha)   # this is the lower one-sided 1-alpha/2 interval 
ucin1=ucit2(xx,n,alpha)   # this is the upper one-sided 1-alpha/2 interval
                          # so [lcin2,ucin2] is a two-sided 1-alpha interval

lciw=lcin1  #  lcin1<=lciw<=lcin2
uciw=ucin1  #  ucin2<=uciw<=ucin1

cpcit<- function(N,lcin,ucin){ 
kk=1:length(N)
for (i in kk){
indp=0:min(n,M)
uu=0 
while (uu<min(n,M)+0.5) {
indp[uu+1]=ind(N[i],lcin[uu+1],ucin[uu+1])*dhyper(uu,M,N[i]-M,n)
uu=uu+1}
kk[i]=sum(indp)
}
kk
}

cpcits<- function(lcin,ucin){  # the minimum of cpcit 
kk=length(lcin)
N11=lcin[1:(kk-1)]-1
N12=ucin[2:kk]+1
N1=union(N11,N12)
ccp=N1
for (i in 1:length(N1)){ #%
indp=0:(kk-1)
uu=0 
while (uu<kk-1+0.5) {
indp[uu+1]=ind(N1[i],lcin[uu+1],ucin[uu+1])*dhyper(uu,M,N1[i]-M,n)
uu=uu+1
}
ccp[i]=sum(indp) 
} #%
min(ccp)
}

#  Need to compute lciw[1] (xvalue=0) since uciw[1]=Inf.

aa=lcin1[1]:lcin2[1]
if (length(aa)>1){ #2
caa=cbind(aa,aa)
for (jj in 1:length(aa)){  #1
lciw[1]=aa[jj]
#N11=lciw[1:min(n,M)]-1
#N12=uciw[2:(min(n,M)+1)]+1
#N1=union(N11,N12)
#caa[jj,2]=min(cpcit(N1,lciw,uciw))
caa[jj,2]=cpcits(lciw,uciw)
} #1
caa=caa[which(caa[,2]>=1-alpha),]
if (length(caa)>2){lciw[1]=caa[dim(caa)[1],1]}
else {lciw[1]=caa[1]}
} #2


# Compute lciw[ii] and uciw[ii] for min(n,M)+1>=ii>=2 when ii goes large
# ii=xvalue+1. xvalue=1 iff ii=2

xvalue=2
while (xvalue<min(n,M)+1.5){ ##
#while (xvalue<2.5){ ##
al=lcin2[xvalue]-lciw[xvalue]+1
ff=cbind(1:al,1:al,1:al,1:al)
ff[,1]=lciw[xvalue]:lcin2[xvalue]
ff[,3]=ff[,3]+Inf

if (length(ff)==4){  #8
upp=ucin2[xvalue]:uciw[xvalue]
bmin=upp[1]
bmax=upp[length(upp)]
while (bmax-bmin>1.5){  #$
bmid=ceiling((bmin+bmax)/2)
uciw[xvalue]=bmid
cpw=cpcits(lciw,uciw)
if (cpw>=1-alpha-10^(-15)){
bmax=bmid} else {bmin=bmid}
} #$
uciw[xvalue]=bmin
cpw=cpcits(lciw,uciw)
if (cpw<1-alpha){uciw[xvalue]=bmax}
} else {            #8
ii=1
while (ii<dim(ff)[1]+0.5)
#for (ii in 1:dim(ff)[1])           
{lciw[xvalue]=ff[ii,1]   #!
upp=ucin2[xvalue]:uciw[xvalue]
bmin=upp[1]
bmax=upp[length(upp)]
uciw[xvalue]=bmax
if (cpcits(lciw,uciw)<1-alpha){ff[ii,3]=Inf
ii=Inf} else { #?
while (bmax-bmin>1.5){  #$
bmid=ceiling((bmin+bmax)/2)
uciw[xvalue]=bmid
cpw=cpcits(lciw,uciw)
if (cpw>=1-alpha-10^(-15)){
bmax=bmid} else {bmin=bmid}
} #$
ff[ii,2]=bmin
uciw[xvalue]=bmin
cpw=cpcits(lciw,uciw)
if (cpw<1-alpha){ff[ii,2]=bmax
uciw[xvalue]=bmax}
ff[ii,3]=ff[ii,2]-ff[ii,1]    
ii=ii+1
} #?
} #!
ff=ff[order(ff[,3]),]
lciw[xvalue]=ff[1,1]
uciw[xvalue]=ff[1,2]
}                  #8
xvalue=xvalue+1
}  ##


N11=lcin1[1:min(n,M)]-1
N12=ucin1[2:(min(n,M)+1)]+1
N1=union(N11,N12)
cp1=cpcit(N1,lcin1,ucin1)

Nw1=lciw[1:min(n,M)]-1
Nw2=uciw[2:(min(n,M)+1)]+1
Nw=union(Nw1,Nw2)
#Nw=max(n,M):10000
cpw=cpcit(Nw,lciw,uciw)

cbind(xx,lcin1,lciw,lcin2,ucin2,uciw,ucin1)
# xx is the all x values, [lcin1,ucin1]=C_O(X), and [lciw,uciw]=C_I(X) in Section 2.3 of the paper.
min(cp1)
min(cpw)


(proc.time()-tim)/60
